/** The AST node for an assignment statement whose left-hand-side is a
 * variable name. */

package AST;

import SymbolTable.*;
import Environment.*;
import IL.*;
import Helpers.*;

public class AssignStmt extends Stmt {
    // Instance variables

    // AST subtrees:
    public String lhs;
    public Expr rhs;

    // The information about the lhs variable, set after typechecking
    protected VarInterface lhs_interface;


    // Constructor
    public AssignStmt(String l, Expr r, int ln) {
	super(ln);
	lhs = l;
	rhs = r;
    }


    // AST printing
    public String toString(String indent) {
	return lhs + " = " + rhs.toString(indent) + ";";
    }


    // Typechecking operations

    /** Typecheck this statement in the given enclosing context */
    public void typecheck(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// look up and remember the information about the lhs variable
	lhs_interface = st.lookupVariable(lhs, line_number);
	
	// extract the type of the lhs variable
	ResolvedType lhs_type = lhs_interface.getType();

	// typecheck the rhs expression
	ResolvedType rhs_type = rhs.typecheck(st);

	// verify that the type of the rhs can be assigned to the type
	// of the lhs
	rhs_type.checkIsAssignableTo(lhs_type, "rhs", "lhs", line_number);
    }


    // Evaluation operations

    /** Evaluate this statement in the given environment */
    public void evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the rhs expression
	Value rhs_value = rhs.evaluate(env);

	// do the assignment
	lhs_interface.assignVariable(env, rhs_value, line_number);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with lowered code
     * for this statement */
    public void lower(ILFunDecl il_fun_decl) {
	// construct the rhs subexpression
	ILExpr rhs_expr = rhs.lower(il_fun_decl);

	// generate the assignment, based on the kind of variable
	// (class or local) we're assigning to
	lhs_interface.generateAssignment(rhs_expr, il_fun_decl);
    }
}
