/** The abstract superclass of all expression AST nodes. */

package AST;

import SymbolTable.*;
import Environment.*;
import IL.*;
import Helpers.*;

public abstract class Expr extends ASTNode {
    // Instance variable

    /** the result type of this expression, computed during
     * typechecking */
    protected ResolvedType result_type;


    // Constructor
    public Expr(int ln) {
	super(ln);
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * remembering the result type, returning the type of the
     * result. */
    public ResolvedType typecheck(CodeSymbolTable t) 
		throws TypecheckCompilerException {
	// typecheck this expression, remembering the result type
	result_type = typecheckHelper(t);
	// return the result type
	return result_type;
    }

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result. Implemented in subclasses. */
    protected abstract ResolvedType typecheckHelper(CodeSymbolTable t) 
	throws TypecheckCompilerException;

    /** Return the result type of this expression, after typechecking */
    public ResolvedType getResultType() {
	if (result_type == null) {
	    throw new InternalCompilerError("asking for the result type " +
					    "before it's been computed");
	}
	return result_type;
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public abstract Value evaluate(CodeEnvironment env) 
	throws EvalCompilerException;


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public abstract ILExpr lower(ILFunDecl il_fun_decl);

    /** Return the lowered result type of this expression, after
     * typechecking */
    public ILType getLoweredResultType() {
	return getResultType().lower();
    }
}
