/** The AST node for a method call expression. */

package AST;

import SymbolTable.*;
import Environment.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class MethodCallExpr extends Expr {
    // Instance variables

    // AST subtrees:
    public Expr receiver;
    public String name;
    public List/*<Expr>*/ args;

    // the interface of the callee method, set after typechecking
    protected MethodInterface method_interface;


    // Constructor
    public MethodCallExpr(Expr r, String n, List/*<Expr>*/ as, int ln) {
	super(ln);
	receiver = r;
	name = n;
	args = as;
    }


    // AST printing
    public String toString(String indent) {
	return 
	    "(" + receiver.toString(indent) + ")." + name + "(" +
	    separatedListToString(args, indent, ", ") + ")";
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result */
    protected ResolvedType typecheckHelper(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// typecheck the receiver expression
	ResolvedType receiver_type = receiver.typecheck(st);

	// typecheck the argument expressions
	List/*<ResolvedType>*/ arg_types = new LinkedList();
	Iterator iter = args.iterator();
	while (iter.hasNext()) {
	    Expr arg = (Expr)iter.next();
	    ResolvedType arg_type = arg.typecheck(st);
	    arg_types.add(arg_type);
	}

	// look up and remember the callee method's interface in the receiver
	method_interface = receiver_type.lookupMethod(name, line_number);

	// check that the actual argument types match the formal
	// argument types

	// get the types of the method's formals
	List/*<ResolvedType>*/ formal_types = 
	    method_interface.getFormalTypes();

	// check if the method is being called with the right
	// number of arguments
	if (formal_types.size() != arg_types.size()) {
	    throw new TypecheckCompilerException(
			"calling method \"" + name +
			"\" with the wrong number of arguments", line_number);
	}

	// check that each actual argument can be assigned to the
	// corresponding formal argument
	Iterator arg_types_iter = arg_types.iterator();
	Iterator formal_types_iter = formal_types.iterator();
	int i = 0;
	while (arg_types_iter.hasNext()) {
	    ResolvedType arg_type = (ResolvedType)arg_types_iter.next();
	    ResolvedType formal_type = (ResolvedType)formal_types_iter.next();
	    arg_type.checkIsAssignableTo(formal_type,
					 "actual #" + i,
					 "formal #" + i, 
					 line_number);
	    i++;
	}

	// all ok!  returns the type of the method's result
	return method_interface.getReturnType();
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the receiver expression
	Value receiver_value = receiver.evaluate(env);

	// get the types of the method's formals
	List/*<ResolvedType>*/ formal_types = 
	    method_interface.getFormalTypes();

	// evaluate the argument expressions, collecting them into a list
	List/*<Value>*/ arg_values = new LinkedList();
	Iterator args_iter = args.iterator();
	Iterator formal_types_iter = formal_types.iterator();
	while (args_iter.hasNext()) {
	    Expr arg = (Expr)args_iter.next();
	    ResolvedType formal_type = (ResolvedType)formal_types_iter.next();

	    // evaluate the argument
	    Value arg_value = arg.evaluate(env);

	    // add the argument to the list
	    arg_values.add(arg_value);
	}

	// look up the method in the receiver
	MethodDeclAndEnv callee = 
	    receiver_value.lookupMethod(name, line_number);

	MethodDecl callee_decl = callee.method_decl;
	ClassEnvironment callee_enclosing_env = callee.enclosing_class_env;

	// evaluate the callee method, and return its result
	return callee_decl.evaluate(receiver_value, arg_values, 
				    callee_enclosing_env, env);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// construct the receiver subexpression
	ILExpr receiver_expr = receiver.lower(il_fun_decl);

	// get the lowered type of the receiver
	ILType lowered_receiver_type = receiver.getLoweredResultType();

	// store it into a variable, so we can refer to it more than once
	ILVar receiver_var = il_fun_decl.newTempVar(lowered_receiver_type);
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(receiver_var),
					     receiver_expr));

	// construct the argument subexpressions
	List/*<ILExpr>*/ arg_exprs = new LinkedList();

	// the receiver is the first (implicit argument)
	arg_exprs.add(new ILVarExpr(receiver_var));

	// get the types of the method's formals
	List/*<ResolvedType>*/ formal_types = 
	    method_interface.getFormalTypes();

	Iterator args_iter = args.iterator();
	Iterator formal_types_iter = formal_types.iterator();
	while (args_iter.hasNext()) {
	    Expr arg = (Expr)args_iter.next();
	    ResolvedType formal_type = (ResolvedType)formal_types_iter.next();

	    // generate code for the argument expression
	    ILExpr arg_expr = arg.lower(il_fun_decl);

	    // add it to the argument list
	    arg_exprs.add(arg_expr);
	}

	// generate a null-pointer check on the receiver
	il_fun_decl.generateNullPtrCheck(new ILVarExpr(receiver_var));

	// get the vtbl record member
	ILType record_type = lowered_receiver_type.getPtrTargetType();
	ILRecordMember vtbl_member = record_type.getRecordMember(".vtbl.");

	// construct code to load the pointer to the vtbl from the
	// vtbl field of the receiver
	ILExpr vtbl_expr = new ILFieldAccessExpr(new ILVarExpr(receiver_var),
						 record_type, vtbl_member);

	// get the method record member
	ILType fun_record_type = vtbl_member.getType().getPtrTargetType();
	ILRecordMember fun_member = fun_record_type.getRecordMember(name);

	// construct code to load the address of the function to call
	// from the vtbl
	ILExpr fun_expr = new ILFieldAccessExpr(vtbl_expr, 
						fun_record_type, fun_member);

	// construct code to call the function indirectly, and return it
	ILType return_type = method_interface.getReturnType().lower();
	return new ILIndirectFunCallExpr(return_type, fun_expr, arg_exprs);
    }
}
