/** The class for the environment for a class, which stores
 * information about each of the class's methods. */

package Environment;

import AST.*;
import SymbolTable.*;
import Helpers.*;
import java.util.*;

public class ClassEnvironment extends NestedEnvironment {
    // Instance variables

    /** the name of the class that this is an environment for */
    protected String class_name;

    /** the environment of the class's superclass, or null if the
     * class doesn't extend any other class */
    protected ClassEnvironment superclass_env;

    /** the instance variables declared in this class and their types
     * (the class doesn't store the values; that's for class instances
     * to store) */
    protected Map/*<String,ResolvedType>*/ instance_variables;

    /** the methods declared in this class */
    protected Map/*<String,MethodDecl>*/ methods;


    // Constructor

    /** create the environment for a class, given the environment for
     * the scope enclosing the class declaration (which is the global
     * environment) */
    public ClassEnvironment(String cn, Environment enclosing_env) {
	super(enclosing_env);
	class_name = cn;
	superclass_env = null;
	instance_variables = new HashMap();
	methods = new HashMap();
    }


    // Accessors

    /** return the name of the class that this is the environment
     * for */
    public String getClassName() {
	return class_name;
    }

    /** record the class's superclass's environmemnt, if the class has a
     * superclass */
    public void setSuperclassEnvironment(ClassEnvironment se) {
	superclass_env = se;
    }

    /** declare a new method in this class */
    public void declareMethod(String name, MethodDecl md, int line_number) {
	if (methods.containsKey(name)) {
	    // we've already declared a method with this same name
	    throw new InternalCompilerError(
		    "redeclaring method named \"" + name + "\"", line_number);
	} else {
	    // remember this method
	    methods.put(name, md);
	}
    }
    
    /** look up the method with the given name in this class or one of
     * its superclasses.  return its declaration and the environment
     * for the class in which it was found. */
    public MethodDeclAndEnv lookupMethod(String name, int line_number) {
	Object o = methods.get(name);
	if (o == null) {
	    // method not found in this class
	    if (superclass_env == null) {
		// and no superclasses to search, so this is a method
		// lookup error
		throw new InternalCompilerError(
		    "calling undefined method \"" + name + "\"", line_number);
	    } else {
		// recursively search the superclass
		return superclass_env.lookupMethod(name, line_number);
	    }
	} else {
	    // method found
	    MethodDecl md = (MethodDecl)o;

	    // return the method along with its enclosing class env
	    return new MethodDeclAndEnv(md, this);
	}
    }

    /** declare a new instance variable in this class */
    public void declareInstanceVariable(ResolvedType type, String name, 
					int line_number) {
	if (instance_variables.containsKey(name)) {
	    // we've already declared an instance variable with this same name
	    throw new InternalCompilerError(
		"redeclaring instance variable named \"" + name + "\"", 
		line_number);
	} else {
	    // remember this instance variable
	    instance_variables.put(name, type);
	}
    }

    /** assign to a previously declared variable */
    public void assignInstanceVariable(Value receiver, String name, 
				       Value value, int line_number) {
	if (instance_variables.containsKey(name)) {
	    // assign to this instance variable, in the receiver object
	    receiver.assignInstanceVariable(name, value, line_number);
	} else {
	    // instance variable not declared in this scope.  but
	    // don't search the lexically enclosing (i.e., global)
	    // scope, instead, search up the inheritance chain.
	    if (superclass_env == null) {
		// there is no superclass; instance variable not declared
		throw new InternalCompilerError(
		    "assigning to undefined variable \"" + name + "\"",
		    line_number);
	    } else {
		// search the superclass
		superclass_env.assignInstanceVariable(receiver, name, value, 
						      line_number);
	    }
	}
    }

    /** look up the current contents of a variable, checking to make
     * sure that we don't access a variable before its been assigned
     * to */
    public Value lookupInstanceVariable(Value receiver, String name, 
					int line_number) {
	if (instance_variables.containsKey(name)) {
	    // read this instance variable, in the receiver object
	    return receiver.lookupInstanceVariable(name, line_number);
	} else {
	    // instance variable not declared in this scope.  but
	    // don't search the lexically enclosing (i.e., global)
	    // scope, instead, search up the inheritance chain.
	    if (superclass_env == null) {
		// there is no superclass; instance variable not declared
		throw new InternalCompilerError(
			"referencing undefined variable \"" + name + "\"",
			line_number);
	    } else {
		// search the superclass
		return superclass_env.lookupInstanceVariable(receiver, name,
							     line_number);
	    }
	}
    }

    /** Initialize all the instance variables in the given instance,
     * which is an instance of the class that this is the environment
     * for */
    public void initializeInstanceVariables(ClassValue instance) {
	Iterator iter = instance_variables.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry entry = (Map.Entry)iter.next();
	    String name = (String)entry.getKey();
	    ResolvedType type = (ResolvedType)entry.getValue();
	    instance.initializeInstanceVariable(name, type);
	}
	if (superclass_env != null) {
	    // also initialize all the inherited instance variables
	    superclass_env.initializeInstanceVariables(instance);
	}
    }


    /** assign to a local variable */
    public void assignLocalVariable(String name, Value value, 
				    int line_number) {
	// no longer in a code environment, so no more local variables
	// to find
	throw new InternalCompilerError(
		    "assigning to undefined local variable \"" + name + "\"",
		    line_number);
    }

    /** look up the current contents of a local variable, checking to
     * make sure that we don't access the variable before it's been
     * assigned to */
    public Value lookupLocalVariable(String name, int line_number)
		throws EvalCompilerException {
	// no longer in a code environment, so no more local variables
	// to find
	throw new InternalCompilerError(
		    "referencing undefined local variable \"" + name + "\"",
		    line_number);
    }
}
