/** This is the class for IL (intermediate language) function
 * declarations */

package IL;

import Target.*;
import Helpers.*;
import java.io.*;
import java.util.*;

public class ILFunDecl {
    // Instance variables

    // the main ones, used to generate target code after the function
    // is fully constructed
    ILType return_type; // the type of its result
    String fun_name; // the name of the function
    List/*<ILFormalVarDecl>*/ formals; // in left-to-right order
    Map/*<String,ILVarDecl>*/ decls; // locals + formals
    List/*<ILStmt>*/ stmts; // in execution order

    // some helpers used only during intermediate-code generation,
    // a.k.a. lowering
    ILProgram program; // the program containing this function
    int var_counter = 0;
    int label_counter = 0;


    // Constructor

    // create a new fun, initially without any formals, locals, or
    // statements
    public ILFunDecl(ILProgram p, ILType rt, String n) { 
	program = p;
	return_type = rt;
	fun_name = n;
	formals = new LinkedList();
	decls = new HashMap();
	stmts = new LinkedList();
    }


    // Basic operations to augment the function with parts not
    // provided during construction

    // declare a new local variable
    public void declareLocal(ILType type, String name) {
	ILLocalVarDecl decl = new ILLocalVarDecl(type, name);
	declareVar(decl);
    }

    // declare a new formal (in left-to-right order)
    public void declareFormal(ILType type, String name) {
	ILFormalVarDecl decl = new ILFormalVarDecl(type, name);
	// formals are recorded two ways
	formals.add(decl);
	declareVar(decl);
    }

    // a helper to declare a new local or formal variable
    protected void declareVar(ILVarDecl decl) {
	String name = decl.getName();
	if (decls.containsKey(name)) {
	    throw new InternalCompilerError(
			"redeclaring local " + name + " in " + fun_name);
	} else {
	    decls.put(name, decl);
	}
    }

    // add a new statement
    public void addStmt(ILStmt stmt) {
	stmts.add(stmt);
    }

    // add a new global var decl to the program containing this function
    public void declareGlobalVar(ILGlobalVarDecl decl) {
	program.addGlobalVarDecl(decl);
    }


    // Accessors

    // return an ILVar for a reference to the given local variable or
    // formal declared in this function
    public ILVar lookupVar(String name) {
	Object d = decls.get(name);
	if (d == null) {
	    throw new InternalCompilerError(
		"referencing undefined local " + name + " in " + fun_name);
	} else {
	    ILVarDecl decl = (ILVarDecl)d;
	    return new ILVar(decl);
	}
    }

    // return an ILGlobalVar for a reference to the given global variable
    public ILGlobalVar lookupGlobalVar(String name) {
	return program.lookupGlobalVar(name);
    }


    // Helper operations for allocating fresh vars and labels

    // return a fresh, new variable name of the given type
    public ILVar newTempVar(ILType type) {
	// create a new temp var whose name is derived from a counter
	// that is incremented each time a new temp is needed
	String name = "T." + Integer.toString(var_counter++);
	// create its declaration
	ILVarDecl decl = new ILLocalVarDecl(type, name);
	// record it and its type
	declareVar(decl);
	// return a reference to it
	return new ILVar(decl);
    }

    // return a fresh, new label
    public ILLabel newLabel() {
	// create a new label whose name is derived from a counter
	// that is incremented each time a new label is needed, plus
	// the name of the enclosing function (to make it globally
	// unique in the output file)
	return new ILLabel("L" + label_counter++ + "." + fun_name);
    }

    // return a fresh, new global variable name of the given type,
    // whose declaration is initialized to the given value
    public ILGlobalVar newInitializedGlobalVar(ILType type, 
					       ILValue initValue) {
	return program.newInitializedGlobalVar(type, initValue);
    }


    // Helper operations for generating standard code sequences

    // generate code to check whether the argument is null, and fail if so
    public void generateNullPtrCheck(ILExpr obj) {
	// construct a comparison of obj != 0 (i.e. null)
	ILExpr test_expr = 
	    new ILIntNotEqualExpr(obj, new ILIntConstantExpr(0));

	// allocate a fresh label for where to go if OK
	ILLabel ok_label = newLabel();
	
	// generate a conditional branch if true
	addStmt(new ILConditionalBranchTrueStmt(test_expr, ok_label));

	// if we fall through to here, then the ptr was null
	addStmt(new ILThrowExceptionStmt("dereferencing a null pointer"));
	// this statement doesn't fall through

	// here's where we should go to if non-null
	addStmt(new ILLabelStmt(ok_label));
    }


    // Printing
    public String toString() {
	String s = return_type.toString() + " " + fun_name + "(";
	Iterator iter = formals.iterator();
	boolean first = true;
	while (iter.hasNext()) {
	    ILFormalVarDecl formal = (ILFormalVarDecl)iter.next();
	    if (first) {
		first = false;
	    } else {
		s += ", ";
	    }
	    s += formal.toString();
	}
	s += ") {\n";
	iter = decls.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry entry = (Map.Entry)iter.next();
	    ILVarDecl var = (ILVarDecl)entry.getValue();
	    if (! var.isFormal()) {
		s += "  " + var.toString() + ";\n";
	    }
	}
	iter = stmts.iterator();
	while (iter.hasNext()) {
	    ILStmt stmt = (ILStmt)iter.next();
	    s += "  " + stmt.toString() + "\n";
	}
	s += "}";
	return s;
    }


    // Target code generation operations

    /** generate code for this program */
    public void codegen(Target target) throws IOException {
	// lay out the function's stack frame, generate whatever the
	// target needs to declare the function, and set up its stack
	// frame
	target.emitFunctionPrologue(return_type, fun_name, formals, decls);

	// generate code for each statement
	Iterator iter = stmts.iterator();
	while (iter.hasNext()) {
	    ILStmt stmt = (ILStmt)iter.next();
	    stmt.codegen(target);
	}

	// generate whatever the target needs at the end of the function
	target.emitFunctionEpilogue(fun_name);
    }
}
