/** The class for IL record types. */

package IL;

import Target.*;
import Helpers.*;
import java.util.*;

public class ILRecordType extends ILType {
    // Instance variables

    // the members of the record type
    protected List/*<ILRecordMember>*/ members;

    // the size of this record, set during code generation
    protected int byteSize = -1; // -1 means not set yet

    // the alignment constraint of this record, set during code generation
    protected int byteAlign = -1; // -1 means not set yet


    // Constructor
    public ILRecordType(List/*<ILRecordMember>*/ ms) {
	members = ms;
    }


    // Accessor

    // if a record type, then return the member with the given name,
    // otherwise report an internal error
    public ILRecordMember getRecordMember(String member_name) {
	Iterator iter = members.iterator();
	while (iter.hasNext()) {
	    ILRecordMember member = (ILRecordMember)iter.next();
	    if (member.getName().equals(member_name)) {
		// found it
		return member;
	    }
	}
	// didn't find it
	throw new InternalCompilerError("didn't find member " + member_name);
    }


    // Printing operations

    // return a printable string representation of this type.
    public String toString() { 
	String s = "{";
	Iterator iter = members.iterator();
	boolean first = true;
	while (iter.hasNext()) {
	    ILRecordMember member = (ILRecordMember)iter.next();
	    if (first) {
		first = false;
	    } else {
		s += ", ";
	    }
	    s += member.toString();
	}
	s += "}";
	return s;
    }


    // Target code generation operations

    /** compute the layout of this type */
    public void doLayout(Target target) {
	// just compute the byteSize, which indirectly will compute
	// the layout
	byteSize(target);
    }	

    /** return the size in bytes of a value of this type on the given
     * target machine */
    public int byteSize(Target target) {
	// the layout of a record is the concatenation of the layout
	// of its members, with each member's starting offset padded
	// out to satisfy its alignment constraint.
	if (byteSize == -1) {
	    // we compute this by computing the starting offset of
	    // each record member in turn.
	    int offset = 0; // we start at offset 0
	    Iterator iter = members.iterator();
	    while (iter.hasNext()) {
		ILRecordMember member = (ILRecordMember)iter.next();
		// pad over the offset based on the necessary alignment
		int memberAlign = member.getType().byteAlign(target);
		offset = Target.align(offset, memberAlign);
		// remember its starting offset, for fast access later
		member.setByteOffset(offset);
		// add its size
		offset += member.getType().byteSize(target);
	    }
	    // the offset after all the members have been added is the
	    // final size
	    byteSize = offset;
	}
	return byteSize;
    }

    /** return the size in bytes of each array element, assuming this
     * type ends with an array type.  report an error if not. */
    public int arrayedElementByteSize(Target target) {
	// the last member must be the arrray
	int size = members.size();
	if (size == 0) {
	    // not arrayed; report an error
	    return super.arrayedElementByteSize(target);
	} else {
	    ILRecordMember last_member = (ILRecordMember)members.get(size-1);
	    return last_member.getType().arrayedElementByteSize(target);
	}
    }

    /** return the alignment constraint in bytes of a value of this
     * type on the given target machine, i.e., at what multiple of
     * bytes must a value of this type begin on */
    public int byteAlign(Target target) {
	if (byteAlign == -1) {
	    // compute the maximum alignment constraint of the members of
	    // the record
	    byteAlign = 1; // an empty record has the minimal alignment
	    Iterator iter = members.iterator();
	    while (iter.hasNext()) {
		ILRecordMember member = (ILRecordMember)iter.next();
		int member_align = member.getType().byteAlign(target);
		if (member_align > byteAlign) byteAlign = member_align;
	    }
	}
	return byteAlign;
    }
}
