/** This is the abstract superclass for all IL (intermediate language)
 * types */

package IL;

import SymbolTable.*;
import Target.*;
import Helpers.*;
import java.util.*;

public abstract class ILType {
    // protected static variable, caching the unique representatives of
    // each distinct class instance record type
    protected static Map/*<ClassSymbolTable,ILRecordType>*/ 
	class_instance_types = new HashMap();

    // protected static variable, caching the unique representatives of
    // each distinct class vtbl record type
    protected static Map/*<ClassSymbolTable,ILRecordType>*/ 
	class_vtbl_types = new HashMap();

    // public static operations, which are the main ones clients
    // should use for getting an IL type
    public static ILIntType intILType() {
	return ILIntType.intILType(); 
    }
    public static ILDoubleType doubleILType() {
	return ILDoubleType.doubleILType(); 
    }
    public static ILFunType funILType() {
	return ILFunType.funILType(); 
    }
    public static ILPtrType ptrILType(ILType target_type) {
	return ILPtrType.ptrILType(target_type); 
    }
    public static ILArrayType arrayILType(ILType elem_type) {
	return ILArrayType.arrayILType(elem_type); 
    }

    /** return the record which is the layout of the virtual function
     * table for this class */
    public static ILRecordType vtblILType(ClassSymbolTable class_st) {
	Object rt = class_vtbl_types.get(class_st);
	if (rt == null) {
	    // class's vtbl record not already computed
	    
	    // construct the members for this vtbl
	    List/*<ILRecordMember>*/ members = new LinkedList();
	    
	    // add fields for all the non-overriding methods of the
	    // class
	    class_st.declareClassVtblRecordMembers(members);

	    // create the record type
	    ILRecordType vrt = new ILRecordType(members);

	    // store the vtbl record into the table
	    class_vtbl_types.put(class_st, vrt);

	    // return it
	    return vrt;
	} else {
	    // already computed
	    return (ILRecordType)rt;
	}
    }

    /** return the record which is the layout of instances of this
     * class */
    public static ILRecordType classILType(ClassSymbolTable class_st) {
	Object rt = class_instance_types.get(class_st);
	if (rt == null) {
	    // class's instance record not already computed
	    
	    // first create an empty placeholder type

	    // construct the members for this instance
	    List/*<ILRecordMember>*/ members = new LinkedList();
	    
	    // create the record type
	    ILRecordType irt = new ILRecordType(members);

	    // store the (empty) instance record into the table (do
	    // this early in case of recursive types)
	    class_instance_types.put(class_st, irt);

	    // now fill in the members of this record

	    // add a member for the class's virtual function table
	    members.add(new ILRecordMember(ptrILType(vtblILType(class_st)),
					   ".vtbl."));

	    // add fields for all the instance variables of the class
	    class_st.declareClassInstanceRecordMembers(members);

	    // return it
	    return irt;
	} else {
	    // already computed
	    return (ILRecordType)rt;
	}
    }


    // Accessors

    // return whether this is the int type
    public boolean isIntType() {
	// by default, it's not
	return false;
    }

    // return whether this is the double type
    public boolean isDoubleType() {
	// by default, it's not
	return false;
    }

    // return whether this is a pointer type
    public boolean isPtrType() {
	// by default, it's not
	return false;
    }

    // if a pointer type, then return the type of the pointer target.
    // otherwise report an internal error
    public ILType getPtrTargetType() {
	throw new InternalCompilerError("expecting a pointer type");
    }

    // if an array type, then return the type of the array elements.
    // otherwise report an internal error
    public ILType getArrayElemType() {
	throw new InternalCompilerError("expecting an array type");
    }

    // if a record type, then return the member with the given name,
    // otherwise report an internal error
    public ILRecordMember getRecordMember(String member_name) {
	throw new InternalCompilerError("expecting a record type");
    }


    // Printing operations

    // return a printable string representation of this type.
    public abstract String toString();

    // return a printable string representation of this type, but
    // don't show target types below pointer types, as a way to break
    // recursion and keep types from getting too big.  typically
    // called only within ILType toString() methods.
    public String shallowToString() {
	// by default, not a pointer type, so treat non-shallowly
	return toString();
    }


    // Target code generation operations

    /** compute the layout of this type */
    public abstract void doLayout(Target target);

    /** return the size in bytes of a value of this type on the given
     * target machine */
    public abstract int byteSize(Target target);

    /** return the size in bytes of each array element, assuming this
     * type ends with an array type.  report an error if not. */
    public int arrayedElementByteSize(Target target) {
	// by default, not arrayed
	throw new InternalCompilerError("not an arrayed type");
    }

    /** return the alignment constraint in bytes of a value of this
     * type on the given target machine, i.e., at what multiple of
     * bytes must a value of this type begin on */
    public abstract int byteAlign(Target target);
}
