/** The class for the interface to a (non-static) method, to store in
 * ClassSymbolTables. */

package SymbolTable;

import java.util.*;

public class MethodInterface {
    // Instance variables

    // the method's return type
    protected ResolvedType return_type;

    // the types of the method's formals, in order
    protected List/*<ResolvedType>*/ formal_types;

    // the root method that this method overrides, transitively (the
    // root method doesn't override any other method), or itself if
    // the method doesn't override any other.  set during typechecking.
    protected MethodInterface root_method;


    // Constructor

    /** create a method interface */
    public MethodInterface(ResolvedType rt, List/*<ResolvedType>*/ fts) {
	return_type = rt;
	formal_types = fts;
    }


    // Accessors

    /** return the method's resolved return type */
    public ResolvedType getReturnType() {
	return return_type;
    }

    /** return the method's ordered list of formal types */
    public List/*<ResolvedType>*/ getFormalTypes() {
	return formal_types;
    }

    /** set the root method that this method overrides,
     * transitively */
    public void setRootMethod(MethodInterface rm) {
	root_method = rm;
    }

    /** return the root method that this method overrides,
     * transitively, or itself if the method doesn't override any
     * other */
    public MethodInterface getRootMethod() {
	return root_method;
    }

    /** return whether this method is a root method, i.e., doesn't
     * override any other */
    public boolean isRootMethod() {
	return root_method == this;
    }


    // Printing operations

    /** return a printable string representation of the method interface */
    public String toString() {
	String s = return_type.toString() + "(";
	Iterator iter = formal_types.iterator();
	boolean first = true;
	while (iter.hasNext()) {
	    ResolvedType formal_type = (ResolvedType)iter.next();
	    if (first) {
		first = false;
	    } else {
		s += ", ";
	    }
	    s += formal_type.toString();
	}
	s += ")";
	return s;
    }


    // Typechecking operations

    /** return whether the receiver properly overrides (i.e., has the
     * same interface as) the argument */
    public boolean overrides(MethodInterface other) {
	// compare return types
	if (! return_type.equals(other.getReturnType())) {
	    // return types don't match
	    return false;
	}

	List/*<ResolvedType>*/ other_formal_types = other.getFormalTypes();

	// compare number of formals
	if (formal_types.size() != other_formal_types.size()) {
	    // differ in number of arguments
	    return false;
	}	

	// compare formal types
	Iterator iter = formal_types.iterator();
	Iterator other_iter = other_formal_types.iterator();
	while (iter.hasNext()) {
	    ResolvedType formal_type = (ResolvedType)iter.next();
	    ResolvedType other_formal_type = (ResolvedType)other_iter.next();
	    if (! formal_type.equals(other_formal_type)) {
		// formal types don't match
		return false;
	    }
	}

	// everything matches
	return true;
    }
}
