/** The class for (resolved) class types.  Only a single instance
 * should ever be created for each distinct class. */

package SymbolTable;

import Environment.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class ResolvedClassType extends ResolvedType {
    // protected static variable, caching the unique representatives of
    // each distinct class type
    protected static Map/*<String,ResolvedClassType>*/ class_types = 
	new HashMap();

    // the symbol table for the class that this type is for
    protected ClassSymbolTable class_interface;

    // Constructor (protected, so clients can't call it directly)
    protected ResolvedClassType(ClassSymbolTable ci) {
	class_interface = ci;
    }

    // public static operation, which is the main one clients should
    // use to get the unique instance of a particular class type
    public static ResolvedClassType classType(String class_name, 
					      SymbolTable st,
					      int line_number)
		throws TypecheckCompilerException { 
	Object rt = class_types.get(class_name);
	if (rt == null) {
	    // class not already resolved
	    
	    // look up the info about the class
	    ClassSymbolTable class_interface = 
		st.lookupClass(class_name, line_number);

	    // create a resolved class type for it
	    ResolvedClassType rct = new ResolvedClassType(class_interface);

	    // store the new resolved type into the table
	    class_types.put(class_name, rct);

	    // return the new resolved class type
	    return rct;
	} else {
	    // resolved class type found
	    return (ResolvedClassType)rt;
	}
    }

    // an alternate version, given a class's symbol table
    public static ResolvedClassType classType(ClassSymbolTable class_st) { 
	String class_name = class_st.getClassName();
	Object rt = class_types.get(class_name);
	if (rt == null) {
	    // class not already resolved

	    // create a resolved class type for it
	    ResolvedClassType rct = new ResolvedClassType(class_st);

	    // store the new resolved type into the table
	    class_types.put(class_name, rct);

	    // return the new resolved class type
	    return rct;
	} else {
	    // resolved class type found
	    return (ResolvedClassType)rt;
	}
    }


    // Printing operations
    public String toString() { 
	return class_interface.getClassName(); 
    }


    // Testers

    /** return whether this type is a class type */
    public boolean isClassType() {
	return true;
    }


    // Accessors

    /** Return the interface to the named class, if a class type */
    public ClassSymbolTable getClassInterface() {
	return class_interface;
    }


    // Type comparison operations

    /** Return whether the receiver should be considered a subtype of
     * the argument */
    public boolean isSubtype(ResolvedType other) {
	if (this == other) {
	    // a type is a subtype of itself
	    return true;
	}
	if (other instanceof ResolvedClassType) {
	    ResolvedClassType other_class_type = (ResolvedClassType)other;
	    // one class type can be a subtype of a different class
	    // type if it inherits from the other
	    return getClassInterface().inheritsFrom(
				    other_class_type.getClassInterface());
	}
	// not subtypes
	return false;
    }
    

    // Typechecking operations

    /** look up the method with the given name in the receiver type.
     * return its interface, or throw an exception if a matching
     * method doesn't exist. */
    public MethodInterface lookupMethod(String name, int line_number)
		throws TypecheckCompilerException {
	// the receiver is a class type, so do the look-up in the
	// class's interface
	MethodInterface mi = class_interface.lookupMethodOrNull(name);
	if (mi == null) {
	    // a matching method wasn't found
	    throw new TypecheckCompilerException(
		    "calling undefined method \"" + name + "\"",
		    line_number);
	} else {
	    // found it
	    return mi;
	}
    }


    // Evaluation operations

    /** Return the default initial value for a variable of this type */
    public Value defaultValue() {
	// null is the default value
	return NullValue.nullValue();
    }


    // Lowering operations

    /** Return the lowered version of this type */
    public ILType lower() {
	// a pointer to a record derived from the class's instance
	// variables
	return ILType.ptrILType(ILType.classILType(class_interface));
    }

    /** Return an IL expression representing the default value of this
     * type */
    public ILExpr defaultILExpr() {
	// null is the default value, which is represented in the
	// lowered code by the integer 0
	return new ILIntConstantExpr(0);
    }

    /** Return an IL value for initializing a global declaration
     * representing the default value of this type */
    public ILValue defaultILValue() {
	// null is the default value, which is represented in the
	// lowered code by the integer 0
	return new ILIntValue(0);
    }
}
